#copyright ReportLab Inc. 2000-2001
#see license.txt for license details
#history http://cvs.sourceforge.net/cgi-bin/cvsweb.cgi/reportlab/graphics/charts/barcharts.py?cvsroot=reportlab
#$Header: /cvsroot/reportlab/reportlab/graphics/charts/barcharts.py,v 1.68 2002/07/31 12:45:08 rgbecker Exp $
"""This module defines a variety of Bar Chart components.

The basic flavors are Side-by-side, available in horizontal and
vertical versions.

Stacked and percentile bar charts to follow...
"""
__version__=''' $Id: barcharts.py,v 1.68 2002/07/31 12:45:08 rgbecker Exp $ '''

import string, copy
from types import FunctionType, StringType

from reportlab.lib import colors
from reportlab.lib.validators import isNumber, isColor, isColorOrNone, isListOfStrings, SequenceOf, isBoolean, isNoneOrShape
from reportlab.lib.formatters import Formatter
from reportlab.lib.attrmap import AttrMap, AttrMapValue
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.graphics.widgetbase import Widget, TypedPropertyCollection, PropHolder
from reportlab.graphics.shapes import Line, Rect, Group, Drawing, NotImplementedError
from reportlab.graphics.charts.axes import XCategoryAxis, YValueAxis
from reportlab.graphics.charts.axes import YCategoryAxis, XValueAxis
from reportlab.graphics.charts.textlabels import BarChartLabel, NA_Label, NoneOrInstanceOfNA_Label
from reportlab.graphics.charts.areas import PlotArea

class BarChartProperties(PropHolder):
    _attrMap = AttrMap(
        strokeColor = AttrMapValue(isColorOrNone, desc='Color of the bar border.'),
        fillColor = AttrMapValue(isColorOrNone, desc='Color of the bar interior area.'),
        strokeWidth = AttrMapValue(isNumber, desc='Width of the bar border.'),
        symbol = AttrMapValue(None, desc='A widget to be used instead of a normal bar.'),
        )

    def __init__(self):
        self.strokeColor = None
        self.fillColor = colors.blue
        self.strokeWidth = 0.5
        self.symbol = None

# Bar chart classes.
class BarChart(PlotArea):
    "Abstract base class, unusable by itself."

    _attrMap = AttrMap(BASE=PlotArea,
        useAbsolute = AttrMapValue(isNumber, desc='Flag to use absolute spacing values.'),
        barWidth = AttrMapValue(isNumber, desc='The width of an individual bar.'),
        groupSpacing = AttrMapValue(isNumber, desc='Width between groups of bars.'),
        barSpacing = AttrMapValue(isNumber, desc='Width between individual bars.'),
        bars = AttrMapValue(None, desc='Handle of the individual bars.'),
        valueAxis = AttrMapValue(None, desc='Handle of the value axis.'),
        categoryAxis = AttrMapValue(None, desc='Handle of the category axis.'),
        data = AttrMapValue(None, desc='Data to be plotted, list of (lists of) numbers.'),
        barLabels = AttrMapValue(None, desc='Handle to the list of bar labels.'),
        barLabelFormat = AttrMapValue(None, desc='Formatting string or function used for bar labels.'),
        reversePlotOrder = AttrMapValue(isBoolean, desc='If true, reverse common category plot order.'),
        naLabel = AttrMapValue(NoneOrInstanceOfNA_Label, desc='Label to use for N/A values.'),
        )

    def __init__(self):
        assert self.__class__.__name__!='BarChart', 'Abstract Class BarChart Instantiated'
        PlotArea.__init__(self)
        self.barSpacing = 0
        self.reversePlotOrder = 0


        # this defines two series of 3 points.  Just an example.
        self.data = [(100,110,120,130),
                    (70, 80, 85, 90)]

        # control bar spacing. is useAbsolute = 1 then
        # the next parameters are in points; otherwise
        # they are 'proportions' and are normalized to
        # fit the available space.  Half a barSpacing
        # is allocated at the beginning and end of the
        # chart.
        self.useAbsolute = 0   #- not done yet
        self.barWidth = 10
        self.groupSpacing = 5
        self.barSpacing = 0

        self.barLabels = TypedPropertyCollection(BarChartLabel)
        self.barLabels.boxAnchor = 'c'
        self.barLabels.textAnchor = 'middle'
        self.barLabelFormat = None

        # this says whether the origin is inside or outside
        # the bar - +10 means put the origin ten points
        # above the tip of the bar if value > 0, or ten
        # points inside if bar value < 0.  This is different
        # to label dx/dy which are not dependent on the
        # sign of the data.
        self.barLabels.nudge = 0

        # if you have multiple series, by default they butt
        # together.

        # we really need some well-designed default lists of
        # colors e.g. from Tufte.  These will be used in a
        # cycle to set the fill color of each series.
        self.bars = TypedPropertyCollection(BarChartProperties)
##        self.bars.symbol = None
        self.bars.strokeWidth = 1
        self.bars.strokeColor = colors.black

        self.bars[0].fillColor = colors.red
        self.bars[1].fillColor = colors.green
        self.bars[2].fillColor = colors.blue
        self.naLabel = None#NA_Label()


    def demo(self):
        """Shows basic use of a bar chart"""
        if self.__class__.__name__=='BarChart':
            raise NotImplementedError, 'Abstract Class BarChart has no demo'
        drawing = Drawing(200, 100)
        bc = self.__class__()
        drawing.add(bc)
        return drawing

    def _getConfigureData(self):
        cA = self.categoryAxis
        data = self.data
        if cA.style!='parallel':
            _data = data
            data = max(map(len,_data))*[0]
            for d in _data:
                for i in xrange(len(d)):
                    data[i] = data[i] + (d[i] or 0)
            data = list(_data) + [data]
        self._configureData = data

    def _getMinMax(self):
        '''Attempt to return the data range'''
        self._getConfigureData()
        self.valueAxis._setRange(self._configureData)
        return self.valueAxis._valueMin, self.valueAxis._valueMax

    def _drawBegin(self,org,length):
        '''Position and configure value axis, return crossing value'''
        vA = self.valueAxis
        vA.setPosition(self.x, self.y, length)
        self._getConfigureData()
        vA.configure(self._configureData)

        # if zero is in chart, put the other axis there, otherwise use low
        crossesAt = vA.scale(0)
        if crossesAt > org+length or crossesAt<org:
            crossesAt = org
        return crossesAt

    def _drawFinish(self):
        '''finalize the drawing of a barchart'''
        cA = self.categoryAxis
        vA = self.valueAxis
        cA.configure(self._configureData)
        self.calcBarPositions()
        g = Group()
        g.add(self.makeBackground())
        cA.makeGrid(g)
        vA.makeGrid(g)
        g.add(self.makeBars())
        g.add(cA)
        g.add(vA)
        del self._configureData
        return g


    def calcBarPositions(self):
        """Works out where they go. default vertical.

        Sets an attribute _barPositions which is a list of
        lists of (x, y, width, height) matching the data.
        """

        flipXY = self._flipXY
        if flipXY:
            org = self.y
        else:
            org = self.x
        cA = self.categoryAxis
        cScale = cA.scale

        data = self.data
        seriesCount = self._seriesCount = len(data)
        self._rowLength = rowLength = max(map(len,data))
        groupSpacing, barSpacing, barWidth = self.groupSpacing, self.barSpacing, self.barWidth
        style = self.categoryAxis.style
        if style=='parallel':
            groupWidth = groupSpacing+(seriesCount*barWidth)+(seriesCount-1)*barSpacing
            bGap = barWidth+barSpacing
        else:
            accum = rowLength*[0]
            groupWidth = groupSpacing+barWidth
            bGap = 0
        self._groupWidth = groupWidth
        useAbsolute = self.useAbsolute

        if useAbsolute:
            # bar dimensions are absolute
            normFactor = 1.0
        else:
            # bar dimensions are normalized to fit.  How wide
            # notionally is one group of bars?
            availWidth = cScale(0)[1]
            normFactor = availWidth/float(groupWidth)
            if self.debug:
                print '%d series, %d points per series' % (seriesCount, self._rowLength)
                print 'width = %d group + (%d bars * %d barWidth) + (%d gaps * %d interBar) = %d total' % (
                    groupSpacing, seriesCount, barWidth,
                    seriesCount-1, barSpacing, groupWidth)

        # 'Baseline' correction...
        vA = self.valueAxis
        vScale = vA.scale
        vm, vM = vA._valueMin, vA._valueMax
        if vm <= 0 <= vM:
            baseLine = vScale(0)
        elif 0 < vm:
            baseLine = vScale(vm)
        elif vM < 0:
            baseLine = vScale(vM)
        self._baseLine = baseLine

        COLUMNS = range(max(map(len,data)))
        if useAbsolute:
            _cScale = cA._scale

        width = self.barWidth*normFactor
        self._barPositions = []
        reversePlotOrder = self.reversePlotOrder
        for rowNo in range(seriesCount):
            barRow = []
            xVal = 0.5*groupSpacing+rowNo*bGap
            if reversePlotOrder: rowNo = seriesCount-1 - rowNo
            for colNo in COLUMNS:
                datum = data[rowNo][colNo]

                # Ufff...
                if useAbsolute:
                    x = groupWidth*_cScale(colNo) + xVal + org
                else:
                    (g, gW) = cScale(colNo)
                    x = g + normFactor*xVal

                if datum is None:
                    height = None
                    y = baseLine
                else:
                    if style!='parallel':
                        y = vScale(accum[colNo])
                        if y<baseLine: y = baseLine
                        accum[colNo] = accum[colNo] + datum
                        datum = accum[colNo]
                    else:
                        y = baseLine
                    height = vScale(datum) - y
                    if -1e-8<height<=1e-8:
                        height = 1e-8
                        if datum<-1e-8: height = -1e-8
                barRow.append(flipXY and (y,x,height,width) or (x, y, width, height))

            self._barPositions.append(barRow)


    def _getLabelText(self, rowNo, colNo):
        '''return formatted label text'''
        labelFmt = self.barLabelFormat
        if labelFmt is None:
            labelText = None
        elif type(labelFmt) is StringType:
            labelText = labelFmt % self.data[rowNo][colNo]
        elif callable(labelFmt):
            labelText = labelFmt(self.data[rowNo][colNo])
        else:
            msg = "Unknown formatter type %s, expected string or function" % labelFmt
            raise Exception, msg
        return labelText

    def _labelXY(self,label,x,y,width,height):
        'Compute x, y for a label'
        if self._flipXY:
            return x + width + (width>=0 and 1 or -1) * label.nudge, y + 0.5*height
        else:
            return x + 0.5*width, y + height + (height>=0 and 1 or -1) * label.nudge

    def _addBarLabel(self, g, rowNo, colNo, x, y, width, height):
        text = self._getLabelText(rowNo,colNo)
        if text:
            self._addLabel(text, self.barLabels[(rowNo, colNo)], g, rowNo, colNo, x, y, width, height)

    def _addNABarLabel(self, g, rowNo, colNo, x, y, width, height):
        na = self.naLabel
        if na and na.text:
            na = copy.copy(na)
            v = self.valueAxis._valueMax<=0 and -1e-8 or 1e-8
            if width is None: width = v
            if height is None: height = v
            self._addLabel(na.text, na, g, rowNo, colNo, x, y, width, height)

    def _addLabel(self, text, label, g, rowNo, colNo, x, y, width, height):
        if label.visible:
            labelWidth = stringWidth(text, label.fontName, label.fontSize)
            x0, y0 = self._labelXY(label,x,y,width,height)
            flipXY = self._flipXY
            if flipXY:
                pm = width
            else:
                pm = height
            label._pmv = pm #the plus minus val
            fixedEnd = getattr(label,'fixedEnd', None)
            if fixedEnd is not None:
                v = fixedEnd._getValue(self,pm)
                x00, y00 = x0, y0
                if flipXY:
                    x0 = v
                else:
                    y0 = v
            else:
                if flipXY:
                    x00 = x0
                    y00 = y+height/2.0
                else:
                    x00 = x+width/2.0
                    y00 = y0
            fixedStart = getattr(label,'fixedStart', None)
            if fixedStart is not None:
                v = fixedStart._getValue(self,pm)
                if flipXY:
                    x00 = v
                else:
                    y00 = v

            if pm<0:
                if flipXY:
                    dx = -2*label.dx
                    dy = 0
                else:
                    dy = -2*label.dy
                    dx = 0
            else:
                dy = dx = 0
            label.setOrigin(x0+dx, y0+dy)
            label.setText(text)
            sC, sW = label.lineStrokeColor, label.lineStrokeWidth
            if sC and sW: g.insert(0,Line(x00,y00,x0,y0, strokeColor=sC, strokeWidth=sW))
            g.add(label)
            alx = getattr(self,'barLabelCallOut',None)
            if alx:
                label._callOutInfo = (self,g,rowNo,colNo,x,y,width,height,x00,y00,x0,y0)
                alx(label)
                del label._callOutInfo

    def makeBars(self):
        g = Group()

        lenData = len(self.data)
        reversePlotOrder = self.reversePlotOrder
        bars = self.bars
        for rowNo in range(lenData):
            if reversePlotOrder: rowNo = lenData-1 - rowNo
            row = self._barPositions[rowNo]
            styleCount = len(bars)
            styleIdx = rowNo % styleCount
            rowStyle = bars[styleIdx]
            for colNo in range(len(row)):
                barPos = row[colNo]
                style = bars.has_key((styleIdx,colNo)) and bars[(styleIdx,colNo)] or rowStyle
                (x, y, width, height) = barPos
                if None in (width,height):
                    self._addNABarLabel(g,rowNo,colNo,x,y,width,height)
                    continue

                # Draw a rectangular symbol for each data item,
                # or a normal colored rectangle.
                symbol = None
                if hasattr(style, 'symbol'):
                    symbol = copy.deepcopy(style.symbol)
                elif hasattr(self.bars, 'symbol'):
                    symbol = self.bars.symbol

                if symbol:
                    symbol.x = x
                    symbol.y = y
                    symbol.width = width
                    symbol.height = height
                    g.add(symbol)
                elif abs(width)>1e-7 and abs(height)>=1e-7 and (style.fillColor is not None or style.strokeColor is not None):
                    r = Rect(x, y, width, height)
                    r.strokeWidth = style.strokeWidth
                    r.fillColor = style.fillColor
                    r.strokeColor = style.strokeColor
                    g.add(r)

                self._addBarLabel(g,rowNo,colNo,x,y,width,height)
        return g

    def _desiredCategoryAxisLength(self):
        '''for dynamically computing the desired category axis length'''
        style = self.categoryAxis.style
        data = self.data
        n = len(data)
        m = max(map(len,data))
        if style=='parallel':
            groupWidth = (n-1)*self.barSpacing+n*self.barWidth
        else:
            groupWidth = self.barWidth
        return m*(self.groupSpacing+groupWidth)

    def draw(self):
        cA, vA = self.categoryAxis, self.valueAxis
        if vA: ovAjA, vA.joinAxis = vA.joinAxis, cA
        if cA: ocAjA, cA.joinAxis = cA.joinAxis, vA
        if self._flipXY:
            cA.setPosition(self._drawBegin(self.x,self.width), self.y, self.height)
        else:
            cA.setPosition(self.x, self._drawBegin(self.y,self.height), self.width)
        return self._drawFinish()

class VerticalBarChart(BarChart):
    "Vertical bar chart with multiple side-by-side bars."

    _flipXY = 0

    def __init__(self):
        self.categoryAxis = XCategoryAxis()
        self.valueAxis = YValueAxis()
        BarChart.__init__(self)

class HorizontalBarChart(BarChart):
    "Horizontal bar chart with multiple side-by-side bars."

    _flipXY = 1

    def __init__(self):
        self.categoryAxis = YCategoryAxis()
        self.valueAxis = XValueAxis()
        BarChart.__init__(self)

# Vertical samples.

def sampleV0a():
    "A slightly pathologic bar chart with only TWO data items."

    drawing = Drawing(400, 200)

    data = [(13, 20)]

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'ne'
    bc.categoryAxis.labels.dx = 8
    bc.categoryAxis.labels.dy = -2
    bc.categoryAxis.labels.angle = 30
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV0b():
    "A pathologic bar chart with only ONE data item."

    drawing = Drawing(400, 200)

    data = [(42,)]

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 50
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'ne'
    bc.categoryAxis.labels.dx = 8
    bc.categoryAxis.labels.dy = -2
    bc.categoryAxis.labels.angle = 30
    bc.categoryAxis.categoryNames = ['Jan-99']

    drawing.add(bc)

    return drawing


def sampleV0c():
    "A really pathologic bar chart with NO data items at all!"

    drawing = Drawing(400, 200)

    data = [()]

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'ne'
    bc.categoryAxis.labels.dx = 8
    bc.categoryAxis.labels.dy = -2
    bc.categoryAxis.categoryNames = []

    drawing.add(bc)

    return drawing


def sampleV1():
    "Sample of multi-series bar chart."

    drawing = Drawing(400, 200)

    data = [
            (13, 5, 20, 22, 37, 45, 19, 4),
            (14, 6, 21, 23, 38, 46, 20, 5)
            ]

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'ne'
    bc.categoryAxis.labels.dx = 8
    bc.categoryAxis.labels.dy = -2
    bc.categoryAxis.labels.angle = 30

    catNames = string.split('Jan Feb Mar Apr May Jun Jul Aug', ' ')
    catNames = map(lambda n:n+'-99', catNames)
    bc.categoryAxis.categoryNames = catNames
    drawing.add(bc)

    return drawing


def sampleV2a():
    "Sample of multi-series bar chart."

    data = [(2.4, -5.7, 2, 5, 9.2),
            (0.6, -4.9, -3, 4, 6.8)
            ]

    labels = ("Q3 2000", "Year to Date", "12 months",
              "Annualised\n3 years", "Since 07.10.99")

    drawing = Drawing(400, 200)

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 120
    bc.width = 300
    bc.data = data

    bc.barSpacing = 0
    bc.groupSpacing = 10
    bc.barWidth = 10

    bc.valueAxis.valueMin = -15
    bc.valueAxis.valueMax = +15
    bc.valueAxis.valueStep = 5
    bc.valueAxis.labels.fontName = 'Helvetica'
    bc.valueAxis.labels.fontSize = 8
    bc.valueAxis.labels.boxAnchor = 'n'   # irrelevant (becomes 'c')
    bc.valueAxis.labels.textAnchor = 'middle'

    bc.categoryAxis.categoryNames = labels
    bc.categoryAxis.labels.fontName = 'Helvetica'
    bc.categoryAxis.labels.fontSize = 8
    bc.categoryAxis.labels.dy = -60

    drawing.add(bc)

    return drawing


def sampleV2b():
    "Sample of multi-series bar chart."

    data = [(2.4, -5.7, 2, 5, 9.2),
            (0.6, -4.9, -3, 4, 6.8)
            ]

    labels = ("Q3 2000", "Year to Date", "12 months",
              "Annualised\n3 years", "Since 07.10.99")

    drawing = Drawing(400, 200)

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 120
    bc.width = 300
    bc.data = data

    bc.barSpacing = 5
    bc.groupSpacing = 10
    bc.barWidth = 10

    bc.valueAxis.valueMin = -15
    bc.valueAxis.valueMax = +15
    bc.valueAxis.valueStep = 5
    bc.valueAxis.labels.fontName = 'Helvetica'
    bc.valueAxis.labels.fontSize = 8
    bc.valueAxis.labels.boxAnchor = 'n'   # irrelevant (becomes 'c')
    bc.valueAxis.labels.textAnchor = 'middle'

    bc.categoryAxis.categoryNames = labels
    bc.categoryAxis.labels.fontName = 'Helvetica'
    bc.categoryAxis.labels.fontSize = 8
    bc.categoryAxis.labels.dy = -60

    drawing.add(bc)

    return drawing


def sampleV2c():
    "Sample of multi-series bar chart."

    data = [(2.4, -5.7, 2, 5, 9.99),
            (0.6, -4.9, -3, 4, 9.99)
            ]

    labels = ("Q3 2000", "Year to Date", "12 months",
              "Annualised\n3 years", "Since 07.10.99")

    drawing = Drawing(400, 200)

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 120
    bc.width = 300
    bc.data = data

    bc.barSpacing = 2
    bc.groupSpacing = 10
    bc.barWidth = 10

    bc.valueAxis.valueMin = -15
    bc.valueAxis.valueMax = +15
    bc.valueAxis.valueStep = 5
    bc.valueAxis.labels.fontName = 'Helvetica'
    bc.valueAxis.labels.fontSize = 8

    bc.categoryAxis.categoryNames = labels
    bc.categoryAxis.labels.fontName = 'Helvetica'
    bc.categoryAxis.labels.fontSize = 8
    bc.valueAxis.labels.boxAnchor = 'n'
    bc.valueAxis.labels.textAnchor = 'middle'
    bc.categoryAxis.labels.dy = -60

    bc.barLabels.nudge = 10

    bc.barLabelFormat = '%0.2f'
    bc.barLabels.dx = 0
    bc.barLabels.dy = 0
    bc.barLabels.boxAnchor = 'n'  # irrelevant (becomes 'c')
    bc.barLabels.fontName = 'Helvetica'
    bc.barLabels.fontSize = 6

    drawing.add(bc)

    return drawing


def sampleV3():
    "Faked horizontal bar chart using a vertical real one (deprecated)."

    names = ("UK Equities", "US Equities", "European Equities", "Japanese Equities",
              "Pacific (ex Japan) Equities", "Emerging Markets Equities",
              "UK Bonds", "Overseas Bonds", "UK Index-Linked", "Cash")

    series1 = (-1.5, 0.3, 0.5, 1.0, 0.8, 0.7, 0.4, 0.1, 1.0, 0.3)
    series2 = (0.0, 0.33, 0.55, 1.1, 0.88, 0.77, 0.44, 0.11, 1.10, 0.33)

    assert len(names) == len(series1), "bad data"
    assert len(names) == len(series2), "bad data"

    drawing = Drawing(400, 200)

    bc = VerticalBarChart()
    bc.x = 0
    bc.y = 0
    bc.height = 100
    bc.width = 150
    bc.data = (series1,)
    bc.bars.fillColor = colors.green

    bc.barLabelFormat = '%0.2f'
    bc.barLabels.dx = 0
    bc.barLabels.dy = 0
    bc.barLabels.boxAnchor = 'w' # irrelevant (becomes 'c')
    bc.barLabels.angle = 90
    bc.barLabels.fontName = 'Helvetica'
    bc.barLabels.fontSize = 6
    bc.barLabels.nudge = 10

    bc.valueAxis.visible = 0
    bc.valueAxis.valueMin = -2
    bc.valueAxis.valueMax = +2
    bc.valueAxis.valueStep = 1

    bc.categoryAxis.tickUp = 0
    bc.categoryAxis.tickDown = 0
    bc.categoryAxis.categoryNames = names
    bc.categoryAxis.labels.angle = 90
    bc.categoryAxis.labels.boxAnchor = 'w'
    bc.categoryAxis.labels.dx = 0
    bc.categoryAxis.labels.dy = -125
    bc.categoryAxis.labels.fontName = 'Helvetica'
    bc.categoryAxis.labels.fontSize = 6

    g = Group(bc)
    g.translate(100, 175)
    g.rotate(-90)

    drawing.add(g)

    return drawing


def sampleV4a():
    "A bar chart showing value axis region starting at *exactly* zero."

    drawing = Drawing(400, 200)

    data = [(13, 20)]

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV4b():
    "A bar chart showing value axis region starting *below* zero."

    drawing = Drawing(400, 200)

    data = [(13, 20)]

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = -10
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV4c():
    "A bar chart showing value axis region staring *above* zero."

    drawing = Drawing(400, 200)

    data = [(13, 20)]

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 10
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV4d():
    "A bar chart showing value axis region entirely *below* zero."

    drawing = Drawing(400, 200)

    data = [(-13, -20)]

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = -30
    bc.valueAxis.valueMax = -10
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


###

##dataSample5 = [(10, 20), (20, 30), (30, 40), (40, 50), (50, 60)]
##dataSample5 = [(10, 60), (20, 50), (30, 40), (40, 30), (50, 20)]
dataSample5 = [(10, 60), (20, 50), (30, 40), (40, 30)]

def sampleV5a():
    "A simple bar chart with no expressed spacing attributes."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV5b():
    "A simple bar chart with proportional spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 0
    bc.barWidth = 40
    bc.groupSpacing = 20
    bc.barSpacing = 10

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV5c1():
    "Make sampe simple bar chart but with absolute spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 1
    bc.barWidth = 40
    bc.groupSpacing = 0
    bc.barSpacing = 0

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV5c2():
    "Make sampe simple bar chart but with absolute spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 1
    bc.barWidth = 40
    bc.groupSpacing = 20
    bc.barSpacing = 0

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV5c3():
    "Make sampe simple bar chart but with absolute spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 1
    bc.barWidth = 40
    bc.groupSpacing = 0
    bc.barSpacing = 10

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleV5c4():
    "Make sampe simple bar chart but with absolute spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 1
    bc.barWidth = 40
    bc.groupSpacing = 20
    bc.barSpacing = 10

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'n'
    bc.categoryAxis.labels.dy = -5
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


# Horizontal samples

def sampleH0a():
    "Make a slightly pathologic bar chart with only TWO data items."

    drawing = Drawing(400, 200)

    data = [(13, 20)]

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'se'
    bc.categoryAxis.labels.angle = 30
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH0b():
    "Make a pathologic bar chart with only ONE data item."

    drawing = Drawing(400, 200)

    data = [(42,)]

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 50
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'se'
    bc.categoryAxis.labels.angle = 30
    bc.categoryAxis.categoryNames = ['Jan-99']

    drawing.add(bc)

    return drawing


def sampleH0c():
    "Make a really pathologic bar chart with NO data items at all!"

    drawing = Drawing(400, 200)

    data = [()]

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'se'
    bc.categoryAxis.labels.angle = 30
    bc.categoryAxis.categoryNames = []

    drawing.add(bc)

    return drawing


def sampleH1():
    "Sample of multi-series bar chart."

    drawing = Drawing(400, 200)

    data = [
            (13, 5, 20, 22, 37, 45, 19, 4),
            (14, 6, 21, 23, 38, 46, 20, 5)
            ]

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    catNames = string.split('Jan Feb Mar Apr May Jun Jul Aug', ' ')
    catNames = map(lambda n:n+'-99', catNames)
    bc.categoryAxis.categoryNames = catNames
    drawing.add(bc)

    return drawing


def sampleH2a():
    "Sample of multi-series bar chart."

    data = [(2.4, -5.7, 2, 5, 9.2),
            (0.6, -4.9, -3, 4, 6.8)
            ]

    labels = ("Q3 2000", "Year to Date", "12 months",
              "Annualised\n3 years", "Since 07.10.99")

    drawing = Drawing(400, 200)

    bc = HorizontalBarChart()
    bc.x = 80
    bc.y = 50
    bc.height = 120
    bc.width = 300
    bc.data = data

    bc.barSpacing = 0
    bc.groupSpacing = 10
    bc.barWidth = 10

    bc.valueAxis.valueMin = -15
    bc.valueAxis.valueMax = +15
    bc.valueAxis.valueStep = 5
    bc.valueAxis.labels.fontName = 'Helvetica'
    bc.valueAxis.labels.fontSize = 8
    bc.valueAxis.labels.boxAnchor = 'n'   # irrelevant (becomes 'c')
    bc.valueAxis.labels.textAnchor = 'middle'
    bc.valueAxis.configure(bc.data)

    bc.categoryAxis.categoryNames = labels
    bc.categoryAxis.labels.fontName = 'Helvetica'
    bc.categoryAxis.labels.fontSize = 8
    bc.categoryAxis.labels.dx = -150

    drawing.add(bc)

    return drawing


def sampleH2b():
    "Sample of multi-series bar chart."

    data = [(2.4, -5.7, 2, 5, 9.2),
            (0.6, -4.9, -3, 4, 6.8)
            ]

    labels = ("Q3 2000", "Year to Date", "12 months",
              "Annualised\n3 years", "Since 07.10.99")

    drawing = Drawing(400, 200)

    bc = HorizontalBarChart()
    bc.x = 80
    bc.y = 50
    bc.height = 120
    bc.width = 300
    bc.data = data

    bc.barSpacing = 5
    bc.groupSpacing = 10
    bc.barWidth = 10

    bc.valueAxis.valueMin = -15
    bc.valueAxis.valueMax = +15
    bc.valueAxis.valueStep = 5
    bc.valueAxis.labels.fontName = 'Helvetica'
    bc.valueAxis.labels.fontSize = 8
    bc.valueAxis.labels.boxAnchor = 'n'   # irrelevant (becomes 'c')
    bc.valueAxis.labels.textAnchor = 'middle'

    bc.categoryAxis.categoryNames = labels
    bc.categoryAxis.labels.fontName = 'Helvetica'
    bc.categoryAxis.labels.fontSize = 8
    bc.categoryAxis.labels.dx = -150

    drawing.add(bc)

    return drawing


def sampleH2c():
    "Sample of multi-series bar chart."

    data = [(2.4, -5.7, 2, 5, 9.99),
            (0.6, -4.9, -3, 4, 9.99)
            ]

    labels = ("Q3 2000", "Year to Date", "12 months",
              "Annualised\n3 years", "Since 07.10.99")

    drawing = Drawing(400, 200)

    bc = HorizontalBarChart()
    bc.x = 80
    bc.y = 50
    bc.height = 120
    bc.width = 300
    bc.data = data

    bc.barSpacing = 2
    bc.groupSpacing = 10
    bc.barWidth = 10

    bc.valueAxis.valueMin = -15
    bc.valueAxis.valueMax = +15
    bc.valueAxis.valueStep = 5
    bc.valueAxis.labels.fontName = 'Helvetica'
    bc.valueAxis.labels.fontSize = 8
    bc.valueAxis.labels.boxAnchor = 'n'
    bc.valueAxis.labels.textAnchor = 'middle'

    bc.categoryAxis.categoryNames = labels
    bc.categoryAxis.labels.fontName = 'Helvetica'
    bc.categoryAxis.labels.fontSize = 8
    bc.categoryAxis.labels.dx = -150

    bc.barLabels.nudge = 10

    bc.barLabelFormat = '%0.2f'
    bc.barLabels.dx = 0
    bc.barLabels.dy = 0
    bc.barLabels.boxAnchor = 'n'  # irrelevant (becomes 'c')
    bc.barLabels.fontName = 'Helvetica'
    bc.barLabels.fontSize = 6

    drawing.add(bc)

    return drawing


def sampleH3():
    "A really horizontal bar chart (compared to the equivalent faked one)."

    names = ("UK Equities", "US Equities", "European Equities", "Japanese Equities",
              "Pacific (ex Japan) Equities", "Emerging Markets Equities",
              "UK Bonds", "Overseas Bonds", "UK Index-Linked", "Cash")

    series1 = (-1.5, 0.3, 0.5, 1.0, 0.8, 0.7, 0.4, 0.1, 1.0, 0.3)
    series2 = (0.0, 0.33, 0.55, 1.1, 0.88, 0.77, 0.44, 0.11, 1.10, 0.33)

    assert len(names) == len(series1), "bad data"
    assert len(names) == len(series2), "bad data"

    drawing = Drawing(400, 200)

    bc = HorizontalBarChart()
    bc.x = 100
    bc.y = 20
    bc.height = 150
    bc.width = 250
    bc.data = (series1,)
    bc.bars.fillColor = colors.green

    bc.barLabelFormat = '%0.2f'
    bc.barLabels.dx = 0
    bc.barLabels.dy = 0
    bc.barLabels.boxAnchor = 'w' # irrelevant (becomes 'c')
    bc.barLabels.fontName = 'Helvetica'
    bc.barLabels.fontSize = 6
    bc.barLabels.nudge = 10

    bc.valueAxis.visible = 0
    bc.valueAxis.valueMin = -2
    bc.valueAxis.valueMax = +2
    bc.valueAxis.valueStep = 1

    bc.categoryAxis.tickLeft = 0
    bc.categoryAxis.tickRight = 0
    bc.categoryAxis.categoryNames = names
    bc.categoryAxis.labels.boxAnchor = 'w'
    bc.categoryAxis.labels.dx = -170
    bc.categoryAxis.labels.fontName = 'Helvetica'
    bc.categoryAxis.labels.fontSize = 6

    g = Group(bc)
    drawing.add(g)

    return drawing


def sampleH4a():
    "A bar chart showing value axis region starting at *exactly* zero."

    drawing = Drawing(400, 200)

    data = [(13, 20)]

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH4b():
    "A bar chart showing value axis region starting *below* zero."

    drawing = Drawing(400, 200)

    data = [(13, 20)]

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = -10
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH4c():
    "A bar chart showing value axis region starting *above* zero."

    drawing = Drawing(400, 200)

    data = [(13, 20)]

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 10
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH4d():
    "A bar chart showing value axis region entirely *below* zero."

    drawing = Drawing(400, 200)

    data = [(-13, -20)]

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data

    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = -30
    bc.valueAxis.valueMax = -10
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


dataSample5 = [(10, 60), (20, 50), (30, 40), (40, 30)]

def sampleH5a():
    "A simple bar chart with no expressed spacing attributes."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH5b():
    "A simple bar chart with proportional spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 0
    bc.barWidth = 40
    bc.groupSpacing = 20
    bc.barSpacing = 10

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH5c1():
    "A simple bar chart with absolute spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 1
    bc.barWidth = 10
    bc.groupSpacing = 0
    bc.barSpacing = 0

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH5c2():
    "Simple bar chart with absolute spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 1
    bc.barWidth = 10
    bc.groupSpacing = 20
    bc.barSpacing = 0

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH5c3():
    "Simple bar chart with absolute spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 20
    bc.height = 155
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 1
    bc.barWidth = 10
    bc.groupSpacing = 0
    bc.barSpacing = 2

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing


def sampleH5c4():
    "Simple bar chart with absolute spacing."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = HorizontalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.useAbsolute = 1
    bc.barWidth = 10
    bc.groupSpacing = 20
    bc.barSpacing = 10

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    drawing.add(bc)

    return drawing

def sampleSymbol1():
    "Simple bar chart using symbol attribute."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = VerticalBarChart()
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.barWidth = 10
    bc.groupSpacing = 15
    bc.barSpacing = 3

    bc.valueAxis.valueMin = 0
    bc.valueAxis.valueMax = 60
    bc.valueAxis.valueStep = 15

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    from reportlab.graphics.widgets.grids import ShadedRect
    sym1 = ShadedRect()
    sym1.fillColorStart = colors.black
    sym1.fillColorEnd = colors.blue
    sym1.orientation = 'horizontal'
    sym1.strokeWidth = 0

    sym2 = ShadedRect()
    sym2.fillColorStart = colors.black
    sym2.fillColorEnd = colors.pink
    sym2.orientation = 'horizontal'
    sym2.strokeWidth = 0

    sym3 = ShadedRect()
    sym3.fillColorStart = colors.blue
    sym3.fillColorEnd = colors.white
    sym3.orientation = 'vertical'
    sym3.cylinderMode = 1
    sym3.strokeWidth = 0

    bc.bars.symbol = sym1
    bc.bars[2].symbol = sym2
    bc.bars[3].symbol = sym3

    drawing.add(bc)

    return drawing

def sampleStacked1():
    "Simple bar chart using symbol attribute."

    drawing = Drawing(400, 200)

    data = dataSample5

    bc = VerticalBarChart()
    bc.categoryAxis.style = 'stacked'
    bc.x = 50
    bc.y = 50
    bc.height = 125
    bc.width = 300
    bc.data = data
    bc.strokeColor = colors.black

    bc.barWidth = 10
    bc.groupSpacing = 15
    bc.valueAxis.valueMin = 0

    bc.categoryAxis.labels.boxAnchor = 'e'
    bc.categoryAxis.categoryNames = ['Ying', 'Yang']

    from reportlab.graphics.widgets.grids import ShadedRect
    bc.bars.symbol = ShadedRect()
    bc.bars.symbol.fillColorStart = colors.red
    bc.bars.symbol.fillColorEnd = colors.white
    bc.bars.symbol.orientation = 'vertical'
    bc.bars.symbol.cylinderMode = 1
    bc.bars.symbol.strokeWidth = 0

    bc.bars[1].symbol = ShadedRect()
    bc.bars[1].symbol.fillColorStart = colors.magenta
    bc.bars[1].symbol.fillColorEnd = colors.white
    bc.bars[1].symbol.orientation = 'vertical'
    bc.bars[1].symbol.cylinderMode = 1
    bc.bars[1].symbol.strokeWidth = 0

    bc.bars[2].symbol = ShadedRect()
    bc.bars[2].symbol.fillColorStart = colors.green
    bc.bars[2].symbol.fillColorEnd = colors.white
    bc.bars[2].symbol.orientation = 'vertical'
    bc.bars[2].symbol.cylinderMode = 1
    bc.bars[2].symbol.strokeWidth = 0

    bc.bars[3].symbol = ShadedRect()
    bc.bars[3].symbol.fillColorStart = colors.blue
    bc.bars[3].symbol.fillColorEnd = colors.white
    bc.bars[3].symbol.orientation = 'vertical'
    bc.bars[3].symbol.cylinderMode = 1
    bc.bars[3].symbol.strokeWidth = 0

    drawing.add(bc)

    return drawing

#class version of function sampleH5c4 above
class SampleH5c4(Drawing):
    "Simple bar chart with absolute spacing."

    def __init__(self,width=400,height=200,*args,**kw):
        apply(Drawing.__init__,(self,width,height)+args,kw)
        bc = HorizontalBarChart()
        bc.x = 50
        bc.y = 50
        bc.height = 125
        bc.width = 300
        bc.data = dataSample5
        bc.strokeColor = colors.black

        bc.useAbsolute = 1
        bc.barWidth = 10
        bc.groupSpacing = 20
        bc.barSpacing = 10

        bc.valueAxis.valueMin = 0
        bc.valueAxis.valueMax = 60
        bc.valueAxis.valueStep = 15

        bc.categoryAxis.labels.boxAnchor = 'e'
        bc.categoryAxis.categoryNames = ['Ying', 'Yang']

        self.add(bc,name='HBC')



#copyright ReportLab Inc. 2000
#see license.txt for license details
#history http://cvs.sourceforge.net/cgi-bin/cvsweb.cgi/reportlab/platypus/frames.py?cvsroot=reportlab
#$Header: /cvsroot/reportlab/reportlab/platypus/frames.py,v 1.18 2002/07/24 19:56:38 andy_robinson Exp $

__version__=''' $Id: frames.py,v 1.18 2002/07/24 19:56:38 andy_robinson Exp $ '''

__doc__="""
"""

_geomAttr=('x1', 'y1', 'width', 'height', 'leftPadding', 'bottomPadding', 'rightPadding', 'topPadding')

_FUZZ=1e-6 # to avoid numerical difficulties we use this as an error band

from reportlab import rl_config

class Frame:
    '''
    A Frame is a piece of space in a document that is filled by the
    "flowables" in the story.  For example in a book like document most
    pages have the text paragraphs in one or two frames.  For generality
    a page might have several frames (for example for 3 column text or
    for text that wraps around a graphic).

    After creation a Frame is not usually manipulated directly by the
    applications program -- it is used internally by the platypus modules.

    Here is a diagramatid abstraction for the definitional part of a Frame

                width                    x2,y2
        +---------------------------------+
        | l  top padding                r | h
        | e +-------------------------+ i | e
        | f |                         | g | i
        | t |                         | h | g
        |   |                         | t | h
        | p |                         |   | t
        | a |                         | p |
        | d |                         | a |
        |   |                         | d |
        |   +-------------------------+   |
        |    bottom padding               |
        +---------------------------------+
        (x1,y1) <-- lower left corner

        NOTE!! Frames are stateful objects.  No single frame should be used in
        two documents at the same time (especially in the presence of multithreading.
    '''
    def __init__(self, x1, y1, width,height, leftPadding=6, bottomPadding=6,
            rightPadding=6, topPadding=6, id=None, showBoundary=0):

        self.id = id

        #these say where it goes on the page
        self.__dict__['_x1'] = x1
        self.__dict__['_y1'] = y1
        self.__dict__['_width'] = width
        self.__dict__['_height'] = height

        #these create some padding.
        self.__dict__['_leftPadding'] = leftPadding
        self.__dict__['_bottomPadding'] = bottomPadding
        self.__dict__['_rightPadding'] = rightPadding
        self.__dict__['_topPadding'] = topPadding

        # if we want a boundary to be shown
        self.showBoundary = showBoundary

        self._geom()
        self._reset()

    def __getattr__(self,a):
        if a in _geomAttr: return self.__dict__['_'+a]
        raise AttributeError, a

    def __setattr__(self,a,v):
        if a in _geomAttr:
            self.__dict__['_'+a] = v
            self._geom()
        else:
            self.__dict__[a] = v

    def _geom(self):
        self._x2 = self._x1 + self._width
        self._y2 = self._y1 + self._height
        #efficiency
        self._y1p = self._y1 + self._bottomPadding
        #work out the available space
        self._aW = self._x2 - self._x1 - self._leftPadding - self._rightPadding
        self._aH = self._y2 - self._y1p - self._topPadding

    def _reset(self):
        #drawing starts at top left
        self._x = self._x1 + self._leftPadding
        self._y = self._y2 - self._topPadding
        self._atTop = 1

    def _add(self, flowable, canv, trySplit=0):
        """ Draws the flowable at the current position.
        Returns 1 if successful, 0 if it would not fit.
        Raises a LayoutError if the object is too wide,
        or if it is too high for a totally empty frame,
        to avoid infinite loops"""
        y = self._y
        p = self._y1p
        s = 0
        if not self._atTop: s = flowable.getSpaceBefore()
        h = y - p - s
        if h>0:
            flowable.canv = canv #so they can use stringWidth etc
            w, h = flowable.wrap(self._aW, h)
            del flowable.canv
        else:
            return 0

        h = h + s
        y = y - h

        if y < p-_FUZZ:
            # width test disabled!  Just allow it to overflow and let the user
            # fix the problem (XXXX MAYBE CHANGE THIS TO OPTIONAL WARNING/ERROR)
            #if ((h > self._aH and not trySplit) or w > self._aW):
            if (not rl_config.allowTableBoundsErrors) and (h > self._aH and not trySplit):
                raise "LayoutError", "Flowable %s (%sx%s points) too large for frame (%sx%s points)." % (
                    flowable.__class__, w,h, self._aW,self._aH)
            return 0
        else:
            #now we can draw it, and update the current point.
            flowable.drawOn(canv, self._x, y, _sW=self._aW-w)
            y = y - flowable.getSpaceAfter()
            if y<>self._y: self._atTop = 0
            self._y = y
            return 1

    add = _add

    def split(self,flowable,canv):
        '''Ask the flowable to split using up the available space.'''
        y = self._y
        p = self._y1p
        s = 0
        if not self._atTop: s = flowable.getSpaceBefore()
        flowable.canv = canv    #some flowables might need this
        r = flowable.split(self._aW, y-p-s)
        del flowable.canv
        return r

    def drawBoundary(self,canv):
        "draw the frame boundary as a rectangle (primarily for debugging)."
        from reportlab.lib.colors import Color, CMYKColor, toColor
        sb = self.showBoundary
        isColor = type(sb) in (type(''),type(()),type([])) or isinstance(sb,Color)
        if isColor:
            sb = toColor(sb,self)
            if sb is self: isColor = 0
            else:
                canv.saveState()
                canv.setStrokeColor(sb)
        canv.rect(
                self._x1,
                self._y1,
                self._x2 - self._x1,
                self._y2 - self._y1
                )
        if isColor: canv.restoreState()

    def addFromList(self, drawlist, canv):
        """Consumes objects from the front of the list until the
        frame is full.  If it cannot fit one object, raises
        an exception."""

        if self.showBoundary:
            self.drawBoundary(canv)

        while len(drawlist) > 0:
            head = drawlist[0]
            if self.add(head,canv,trySplit=0):
                del drawlist[0]
            else:
                #leave it in the list for later
                break


#copyright ReportLab Inc. 2000
#see license.txt for license details
#history http://cvs.sourceforge.net/cgi-bin/cvsweb.cgi/reportlab/demos/odyssey/fodyssey.py?cvsroot=reportlab
#$Header: /cvsroot/reportlab/reportlab/demos/odyssey/fodyssey.py,v 1.17 2002/07/24 19:56:35 andy_robinson Exp $
__version__=''' $Id: fodyssey.py,v 1.17 2002/07/24 19:56:35 andy_robinson Exp $ '''
__doc__=''

#REPORTLAB_TEST_SCRIPT
import sys, copy, string, os
from reportlab.platypus import *
from reportlab.lib.units import inch
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.enums import TA_LEFT, TA_RIGHT, TA_CENTER, TA_JUSTIFY

styles = getSampleStyleSheet()

Title = "The Odyssey"
Author = "Homer"

def myFirstPage(canvas, doc):
    canvas.saveState()
    canvas.restoreState()

def myLaterPages(canvas, doc):
    canvas.saveState()
    canvas.setFont('Times-Roman',9)
    canvas.drawString(inch, 0.75 * inch, "Page %d" % doc.page)
    canvas.restoreState()

def go():
    doc = SimpleDocTemplate('fodyssey.pdf',showBoundary='showboundary' in sys.argv)
    doc.allowSplitting = not 'nosplitting' in sys.argv
    doc.build(Elements,myFirstPage,myLaterPages)

Elements = []

ChapterStyle = copy.copy(styles["Heading1"])
ChapterStyle.alignment = TA_CENTER
ChapterStyle.fontsize = 16
InitialStyle = copy.deepcopy(ChapterStyle)
InitialStyle.fontsize = 16
InitialStyle.leading = 20
PreStyle = styles["Code"]

def newPage():
    Elements.append(PageBreak())

def chapter(txt, style=ChapterStyle):
    newPage()
    Elements.append(Paragraph(txt, style))
    Elements.append(Spacer(0.2*inch, 0.3*inch))

def fTitle(txt,style=InitialStyle):
    Elements.append(Paragraph(txt, style))

ParaStyle = copy.deepcopy(styles["Normal"])
ParaStyle.spaceBefore = 0.1*inch
if 'right' in sys.argv:
    ParaStyle.alignment = TA_RIGHT
elif 'left' in sys.argv:
    ParaStyle.alignment = TA_LEFT
elif 'justify' in sys.argv:
    ParaStyle.alignment = TA_JUSTIFY
elif 'center' in sys.argv or 'centre' in sys.argv:
    ParaStyle.alignment = TA_CENTER
else:
    ParaStyle.alignment = TA_JUSTIFY

def spacer(inches):
    Elements.append(Spacer(0.1*inch, inches*inch))

def p(txt, style=ParaStyle):
    Elements.append(Paragraph(txt, style))

def pre(txt, style=PreStyle):
    spacer(0.1)
    p = Preformatted(txt, style)
    Elements.append(p)

def parseOdyssey(fn):
    from time import time
    E = []
    t0=time()
    L = open(fn,'r').readlines()
    t1 = time()
    print "open(%s,'r').readlines() took %.4f seconds" %(fn,t1-t0)
    for i in xrange(len(L)):
        if L[i][-1]=='\012':
            L[i] = L[i][:-1]
    t2 = time()
    print "Removing all linefeeds took %.4f seconds" %(t2-t1)
    L.append('')
    L.append('-----')

    def findNext(L, i):
        while 1:
            if string.strip(L[i])=='':
                del L[i]
                kind = 1
                if i<len(L):
                    while string.strip(L[i])=='':
                        del L[i]

                    if i<len(L):
                        kind = L[i][-1]=='-' and L[i][0]=='-'
                        if kind:
                            del L[i]
                            if i<len(L):
                                while string.strip(L[i])=='':
                                    del L[i]
                break
            else:
                i = i + 1

        return i, kind

    f = s = 0
    while 1:
        f, k = findNext(L,0)
        if k: break

    E.append([spacer,2])
    E.append([fTitle,'<font color=red>%s</font>' % Title, InitialStyle])
    E.append([fTitle,'<font size=-4>by</font> <font color=green>%s</font>' % Author, InitialStyle])

    while 1:
        if f>=len(L): break

        if string.upper(L[f][0:5])=='BOOK ':
            E.append([chapter,L[f]])
            f=f+1
            while string.strip(L[f])=='': del L[f]
            style = ParaStyle
            func = p
        else:
            style = PreStyle
            func = pre

        while 1:
            s=f
            f, k=findNext(L,s)
            sep= (func is pre) and '\012' or ' '
            E.append([func,string.join(L[s:f],sep),style])
            if k: break
    t3 = time()
    print "Parsing into memory took %.4f seconds" %(t3-t2)
    del L
    t4 = time()
    print "Deleting list of lines took %.4f seconds" %(t4-t3)
    for i in xrange(len(E)):
        apply(E[i][0],E[i][1:])
    t5 = time()
    print "Moving into platypus took %.4f seconds" %(t5-t4)
    del E
    t6 = time()
    print "Deleting list of actions took %.4f seconds" %(t6-t5)
    go()
    t7 = time()
    print "saving to PDF took %.4f seconds" %(t7-t6)
    print "Total run took %.4f seconds"%(t7-t0)

for fn in ('odyssey.full.txt','odyssey.txt'):
    if os.path.isfile(fn):
        break
if __name__=='__main__':
    parseOdyssey(fn)


#copyright ReportLab Inc. 2000-2001
#see license.txt for license details
#history http://cvs.sourceforge.net/cgi-bin/cvsweb.cgi/reportlab/rl_config.py?cvsroot=reportlab
#$Header: /cvsroot/reportlab/reportlab/rl_config.py,v 1.34 2002/07/24 19:56:34 andy_robinson Exp $
__version__=''' $Id: rl_config.py,v 1.34 2002/07/24 19:56:34 andy_robinson Exp $ '''

allowTableBoundsErrors = 1 # set to 0 to die on too large elements in tables in debug (recommend 1 for production use)
shapeChecking =             1
defaultEncoding =           'WinAnsiEncoding'       # 'WinAnsi' or 'MacRoman'
pageCompression =           1                       # default page compression mode
defaultPageSize =           'A4'                    #default page size
defaultImageCaching =       0                       #set to zero to remove those annoying cached images
PIL_WARNINGS =              1                       #set to zero to remove those annoying warnings
ZLIB_WARNINGS =             1
warnOnMissingFontGlyphs =   0                       #if 1, warns of each missing glyph
verbose =                   0
showBoundary =              0                       # turns on and off boundary behaviour in Drawing
emptyTableAction=           'error'                 # one of 'error', 'indicate', 'ignore'

# places to look for T1Font information
T1SearchPath =  ('/usr/local/lib/Acrobat5/Resource/Font',
		 '/usr/local/lib/Acrobat4/Resource/Font',
		 '%(REPORTLAB_DIR)s/fonts' #special
		)

# places to look for TT Font information
TTFSearchPath = ('/usr/X11R6/lib/X11/fonts/TrueType/',
		 '%(REPORTLAB_DIR)s/fonts' #special
		)

# places to look for CMap files - should ideally merge with above
CMapSearchPath = ('/usr/local/Acrobat5/Resource/CMap',
		  '/usr/local/Acrobat4/Resource/CMap',
		  '/usr/local/share/ghostscript/Resource/CMap'
		 )

#### Normally don't need to edit below here ####
import os, sys, string
from reportlab.lib import pagesizes

def _setOpt(name, value, conv=None):
    '''set a module level value from environ/default'''
    from os import environ
    ename = 'RL_'+name
    if environ.has_key(ename):
        value = environ[ename]
    if conv: value = conv(value)
    globals()[name] = value

sys_version = string.split(sys.version)[0]      #strip off the other garbage
_SAVED = {}

def _startUp():
    '''This function allows easy resetting to the global defaults
    If the environment contains 'RL_xxx' then we use the value
    else we use the given default'''
    V = ('T1SearchPath','CMapSearchPath', 'TTFSearchPath',
                'shapeChecking', 'defaultEncoding',
                'pageCompression', 'defaultPageSize', 'defaultImageCaching', 'PIL_WARNINGS',
                'ZLIB_WARNINGS', 'warnOnMissingFontGlyphs', 'verbose', 'emptyTableAction',
                )

    if _SAVED=={}:
        for k in V:
            _SAVED[k] = globals()[k]

    #places to search for Type 1 Font files
    import reportlab
    D = {'REPORTLAB_DIR': os.path.abspath(os.path.dirname(reportlab.__file__)),
        'disk': string.split(os.getcwd(), ':')[0],
        'sys_version': sys_version,
        }

    for name in ('T1SearchPath','TTFSearchPath','CMapSearchPath'):
        P=[]
        for p in _SAVED[name]:
            d = string.replace(p % D,'/',os.sep)
            if os.path.isdir(d): P.append(d)
        _setOpt(name,P)

    for k in V[3:]:
        v = _SAVED[k]
        if type(v)==type(1): conv = int
        elif k=='defaultPageSize': conv = lambda v,M=pagesizes: getattr(M,v)
        else: conv = None
        _setOpt(k,v,conv)

_startUp()



#!/usr/bin/env python

""" Python Highlighter for PDF                     Version: 0.6

    py2pdf.py [options] <file1> [<file2> ...]

    options:
     -h  or  --help    print help (this message)
     -                 read from stdin (writes to stdout)
     --stdout          read from file, write to stdout
                         (restricted to first file only)
     --title=<title>   specify title
     --config=<file>   read configuration options from <file>
     --input=<type>    set input file type
                         'python': Python code (default)
                         'ascii':  arbitrary ASCII files (b/w)
     --mode=<mode>     set output mode
                         'color': output in color (default)
                         'mono':  output in b/w
     --paperFormat=    set paper format (ISO A, B, C series,
       <format>          US legal & letter; default: 'A4')
                         e.g. 'letter', 'A3', 'A4', 'B5', 'C6', ...
     --paperSize=      set paper size in points (size being a valid
       <size>          numeric 2-tuple (x,y) w/o any whitespace)
     --landscape       set landscape format (default: portrait)
     --bgCol=<col>     set page background-color in hex code like
                         '#FFA024' or '0xFFA024' for RGB components
                         (overwrites mono mode)
     --<cat>Col=<col>  set color of certain code categories, i.e.
                         <cat> can be the following:
                         'comm':  comments
                         'ident': identifiers
                         'kw':    keywords
                         'strng': strings
                         'param': parameters
                         'rest':  all the rest
     --fontName=<name> set base font name (default: 'Courier')
                         like 'Helvetica', 'Times-Roman'
     --fontSize=<size> set font size (default: 8)
     --tabSize=<size>  set tab size (default: 4)
     --lineNum         print line numbers
     --multiPage       generate one file per page (with filenames
                         tagged by 1, 2...), disables PDF outline
     --noOutline       don't generate PDF outline (default: unset)
     -v  or  --verbose set verbose mode

    Takes the input, assuming it is Python source code and formats
    it into PDF.

    * Uses Just van Rossum's PyFontify version 0.4 to tag Python
      scripts, now included in reportlab.lib.

    * Uses the ReportLab library (version 0.92 or higher) to
      generate PDF. You can get it without charge from ReportLab:
          http://www.reportlab.com

    * Parts of this code still borrow heavily from Marc-Andre
      Lemburg's py2html who has kindly given permission to
      include them in py2pdf. Thanks, M.-A.!
"""

__copyright__ = """
----------------------------------------------------------------------
(c) Copyright by Dinu C. Gherman, 2000  (gherman@europemail.com)

    Permission to use, copy, modify, and distribute this
    software and its documentation for any purpose and
    without fee or royalty is hereby granted, provided
    that the above copyright notice appear in all copies
    and that both that copyright notice and this permission
    notice appear in supporting documentation or portions
    thereof, including modifications, that you make.

    THE AUTHOR DINU C. GHERMAN DISCLAIMS ALL WARRANTIES
    WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
    SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
    OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
    RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
    IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
    ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
    OR PERFORMANCE OF THIS SOFTWARE!
"""

__version__ = '0.6'
__author__  = 'Dinu C. Gherman'
__date__    = '2001-08-17'
__url__     = 'http://starship.python.net/crew/gherman/programs/py2pdf'


import sys, string, re, os, getopt

from reportlab.pdfgen import canvas
from reportlab.lib.colors import Color, HexColor
from reportlab.lib import fonts
from reportlab.lib.units import cm, inch


### Helpers functions.

def makeTuple(aString):
    """Evaluate a string securely into a tuple.

    Match the string and return an evaluated object thereof if and
    only if we think it is a tuple of two or more numeric decimal(!)
    values, integers or floats. E-notation, hex or octal is not
    supported, though! Shorthand notation (omitting leading or trail-
    zeros, before or after the decimal point) like .25 or 25. is
    supported.
    """

    c = string.count(aString, ',')
    num = '(\d*?\.?\d+?)|(\d+?\.?\d*?)'
    tup = string.join([num]*c, ',')

    if re.match('\(' + tup + '\)', aString):
        return eval(aString)
    else:
        details = '%s cannot be parsed into a numeric tuple!' % aString
        raise 'ValueError', details


def loadFontifier(options=None):
    "Load a tagging module and return a corresponding function."

    # We can't use 'if options' because of the modified
    # attribute lookup it seems.

    if type(options) != type(None) and options.marcs:
        # Use mxTextTool's tagging engine.

        from mxTextTools import tag
        from mxTextTools.Examples.Python import python_script
        tagFunc = lambda text, tag=tag, pytable=python_script: \
                  tag(text,pytable)[1]

    else:
        # Load Just's.

        try:
            from reportlab.lib import PyFontify

            if PyFontify.__version__ < '0.3':
                raise ValueError

            tagFunc = PyFontify.fontify

        except:
            print """
    Sorry, but this script needs the PyFontify.py module version 0.3
    or higher; You can download it from Just's homepage at

       URL: http://starship.python.net/crew/just
"""
            sys.exit()

    return tagFunc


def makeColorFromString(aString):
    """Convert a string to a ReportLab RGB color object.

    Supported formats are: '0xFFFFFF', '#FFFFFF' and '(R,G,B)'
    where the latter is a tuple of three floats in the range
    [0.0, 1.0].
    """

    s = aString

    if s[0] == '#' or s[:2] in ('0x', '0X'):
        if s[:2] in ('0x', '0X'):
            return HexColor('#' + s[2:])

    elif s[0] == '(':
        r, g, b = makeTuple(aString)
        return Color(r, g, b)


### Utility classes.

# For py2pdf this is some kind of overkill, but it's fun, too.
class PaperFormat:
    """Class to capture a paper format/size and its orientation.

    This class represents an abstract paper format, including
    the size and orientation of a sheet of paper in some formats
    plus a few operations to change its size and orientation.
    """

    _A4W, _A4H = 21*cm, 29.7*cm
    A0 = (4*_A4W, 4*_A4H)

    _B4W, _B4H = 25*cm, 35.3*cm
    B0 = (4*_B4W, 4*_B4H)

    _C4W, _C4H = 22.9*cm, 32.4*cm
    C0 = (4*_C4W, 4*_C4H)

    letter = (8.5*inch, 11*inch)
    legal = (8.5*inch, 14*inch)


    def __init__(self, nameOrSize='A4', landscape=0):
        "Initialisation."

        t = type(nameOrSize)

        if t == type(''):
            self.setFormatName(nameOrSize, landscape)
        elif t == type((1,)):
            self.setSize(nameOrSize)
            self.setLandscape(landscape)


    def __repr__(self):
        """Return a string representation of ourself.

        The returned string can also be used to recreate
        the same PaperFormat object again.
        """

        if self.name != 'custom':
            nos = `self.name`
        else:
            nos = `self.size`

        format = "PaperFormat(nameOrSize=%s, landscape=%d)"
        tuple = (nos, self.landscape)

        return format % tuple


    def setSize(self, size=None):
        "Set explicit paper size."

        self.name = 'custom'
        x, y = self.size = size
        self.landscape = x < y


    def setLandscape(self, flag):
        "Set paper orientation as desired."

        # Swap paper orientation if needed.

        self.landscape = flag
        x, y = self.size

        ls = self.landscape
        if (ls and x < y) or (not ls and x > y):
            self.size = y, x


    def setFormatName(self, name='A4', landscape=0):
        "Set paper size derived from a format name."

        if name[0] in 'ABC':
            # Assume ISO-A, -B, -C series.
            # (Hmm, are B and C really ISO standards
            # or just DIN? Well...)
            c, f = name[0], int(name[1:])
            self.size = getattr(self, c + '0')
            self.name = c + '0'
            self.makeHalfSize(f)

        elif name == 'letter':
            self.size = self.letter
            self.name = 'letter'

        elif name == 'legal':
            self.size = self.legal
            self.name = 'legal'

        self.setLandscape(landscape)


    def makeHalfSize(self, times=1):
        "Reduce paper size (surface) by 50% multiple times."

        # Orientation remains unchanged.

        # Iterates only for times >= 1.
        for i in xrange(times):
            s = self.size
            self.size = s[1] / 2.0, s[0]

            if self.name[0] in 'ABC':
                self.name = self.name[0] + `int(self.name[1:]) + 1`
            else:
                self.name = 'custom'


    def makeDoubleSize(self, times=1):
        "Increase paper size (surface) by 50% multiple times."

        # Orientation remains unchanged.

        # Iterates only for times >= 1.
        for i in xrange(times):
            s = self.size
            self.size = s[1], s[0] * 2.0

            if self.name[0] in 'ABC':
                self.name = self.name[0] + `int(self.name[1:]) - 1`
            else:
                self.name = 'custom'


class Options:
    """Container class for options from command line and config files.

    This class is a container for options as they are specified
    when a program is called from a command line, but also from
    the same kind of options that are saved in configuration
    files. Both the short UNIX style (e.g. '-v') as well as the
    extended GNU style (e.g. '--help') are supported.

    An 'option' is a <name>/<value> pair with both parts being
    strings at the beginning, but where <value> might be conver-
    ted later into any other Python object.

    Option values can be accessed by using their name as an attri-
    bute of an Options object, returning None if no such option
    name exists (that makes None values impossible, but so what?).
    """

    # Hmm, could also use UserDict... maybe later.

    def __init__(self):
        "Initialize with some default options name/values."

        # Hmm, could also pass an initial optional dict...

        # Create a dictionary containing the options
        # and populate it with defaults.
        self.pool = {}
        self.setDefaults()


    def __getattr__(self, name):
        "Turn attribute access into dictionary lookup."

        return self.pool.get(name)


    def setDefaults(self):
        "Set default options."

        ### Maybe get these from a site config file...
        self.pool.update({'fontName' : 'Courier',
            'fontSize'   : 8,
            'bgCol'      : Color(1, 1, 1),
            'mode'       : 'color',
            'lineNum'    : 0,
            'tabSize'    : 4,
            'paperFormat': 'A4',
            'landscape'  : 0,
            'title'      : None,
            'multiPage'  : 0})

        # Default colors (for color mode), mostly taken from py2html.
        self.pool.update({'commCol' : HexColor('#1111CC'),
            'kwCol'    : HexColor('#3333CC'),
            'identCol' : HexColor('#CC0000'),
            'paramCol' : HexColor('#000066'),
            'strngCol' : HexColor('#119911'),
            'restCol'  : HexColor('#000000')})

        # Add a default 'real' paper format object.
        pf = PaperFormat(self.paperFormat, self.landscape)
        self.pool.update({'realPaperFormat' : pf})
        self.pool.update({'files' : []})


    def display(self):
        "Display all current option names and values."

        self.saveToFile(sys.stdout)


    def saveToFile(self, path):
        "Save options as a log file."

        if type(path) == type(''):
            f = open(path, 'w')
        else:
            # Assume a file-like object.
            f = path

        items = self.pool.items()
        items.sort()

        for n, v in items:
            f.write("%-15s : %s\n" % (n, `v`))


    def updateWithContentsOfFile(self, path):
        """Update options as specified in a config file.

        The file is expected to contain one option name/value pair
        (seperated by an equal sign) per line, but no other whitespace.
        Option values may contain equal signs, but no whitespace.
        Option names must be valid Python strings, preceeded by one
        or two dashes.
        """

        config = open(path).read()
        config = string.split(config, '\n')

        for cfg in config:
            if cfg == None:
                break

            if cfg == '' or cfg[0] == '#':
                continue

            # GNU long options
            if '=' in cfg and cfg[:2] == '--':
                p = string.find(cfg, '=')
                opt, arg = cfg[2:p], cfg[p+1:]
                self.updateOption(opt, arg)

            # Maybe treat single-letter options as well?
            # elif ':' in cfg and cfg[0] == '-' and cfg[1] != '-':
            #    pass

            else:
                self.updateOption(cfg[2:], None)


    def updateWithContentsOfArgv(self, argv):
        "Update options as specified in a (command line) argument vector."

        # Specify accepted short option names (UNIX style).
        shortOpts = 'hv'

        # Specify accepted long option names (GNU style).
        lo = 'tabSize= paperFormat= paperSize= landscape stdout title= fontName= fontSize='
        lo = lo + ' bgCol= verbose lineNum marcs help multiPage noOutline config= input= mode='
        lo = lo + ' commCol= identCol= kwCol= strngCol= paramCol= restCol='
        longOpts = string.split(lo, ' ')

        try:
            optList, args = getopt.getopt(argv, shortOpts, longOpts)
        except getopt.error, msg:
            sys.stderr.write("%s\nuse -h or --help for help\n" % str(msg))
            sys.exit(2)

        self.updateOption('files', args) # Hmm, really needed?

        for o, v in optList:
            # Remove leading dashes (max. two).
            if o[0] == '-':
                o = o[1:]
            if o[0] == '-':
                o = o[1:]

            self.updateOption(o, v)


    def updateOption(self, name, value):
        "Update an option from a string value."

        # Special treatment for coloring options...
        if name[-3:] == 'Col':
            if name[:-3] in string.split('bg comm ident kw strng param rest', ' '):
                self.pool[name] = makeColorFromString(value)

        elif name == 'paperSize':
            tup = makeTuple(value)
            self.pool['paperSize'] = tup
            if not self.realPaperFormat:
                pf = PaperFormat(self.paperFormat, self.landscape)
                self.pool['realPaperFormat'] = pf
            self.pool['realPaperFormat'].setSize(tup)

        elif name == 'paperFormat':
            self.pool['paperFormat'] = value
            if not self.realPaperFormat:
                pf = PaperFormat(self.paperFormat, self.landscape)
                self.pool['realPaperFormat'] = pf
            self.pool['realPaperFormat'].setFormatName(self.paperFormat, self.landscape)

        elif name == 'landscape':
            self.pool['landscape'] = 1
            if self.realPaperFormat:
                self.pool['realPaperFormat'].setLandscape(1)

        elif name == 'fontSize':
            self.pool['fontSize'] = int(value)

        elif name == 'tabSize':
            self.pool['tabSize'] = int(value)

        elif name == 'mode':
            self.pool['mode'] = value
            if value == 'mono':
                cats = 'comm ident kw strng param rest'
                for cat in string.split(cats, ' '):
                    self.pool[cat + 'Col'] = Color(0, 0, 0)

        # Parse configuration file...
        elif name == 'config':
            self.updateWithContentsOfFile(value)

        elif name == 'stdout':
            self.pool['stdout'] = 1

        elif name == 'files':
            self.pool['files'] = value

        else:
            # Set the value found or 1 for options without values.
            self.pool[name] = value or 1


    def update(self, **options):
        "Update options."

        # Not much tested and/or used, yet!!

        for n, v in options.items():
            self.pool[n] = v


### Layouting classes.

class PDFLayouter:
    """A class to layout a simple PDF document.

    This is intended to help generate PDF documents where all pages
    follow the same kind of 'template' which is supposed to be the
    same adornments (header, footer, etc.) on each page plus a main
    'frame' on each page. These frames are 'connected' such that one
    can add individual text lines one by one with automatic line
    wrapping, page breaks and text flow between frames.
    """

    def __init__(self, options):
        "Initialisation."

        self.options = options
        self.canvas = None
        self.multiLineStringStarted = 0
        self.lineNum = 0

        # Set a default color and font.
        o = self.options
        self.currColor = o.restCol
        self.currFont = (o.fontName, o.fontSize)


    ### Helper methods.

    def setMainFrame(self, frame=None):
        "Define the main drawing frame of interest for each page."

        if frame:
            self.frame = frame
        else:
            # Maybe a long-term candidate for additional options...
            width, height = self.options.realPaperFormat.size
            self.frame = height - 3*cm, 3*cm, 2*cm, width - 2*cm

            # self.frame is a 4-tuple:
            # (topMargin, bottomMargin, leftMargin, rightMargin)


    def setPDFMetaInfo(self):
        "Set PDF meta information."

        o = self.options
        c = self.canvas
        c.setAuthor('py2pdf %s' % __version__)
        c.setSubject('')

        # Set filename.
        filename = ''

        # For stdin use title option or empty...
        if self.srcPath == sys.stdin:
            if o.title:
                filename = o.title
        # otherwise take the input file's name.
        else:
            path = os.path.basename(self.srcPath)
            filename = o.title or path

        c.setTitle(filename)


    def setFillColorAndFont(self, color, font):
        "Set new color/font (maintaining the current 'state')."

        self.currFont = font
        self.currColor = color

        fontName, fontSize = font
        self.text.setFont(fontName, fontSize)
        self.text.setFillColor(color)


    ### API

    def begin(self, srcPath, numLines):
        "Things to do before doing anything else."

        self.lineNum = 0
        self.pageNum = 0
        self.numLines = numLines
        self.srcPath = srcPath

        # Set output filename (stdout if desired).
        o = self.options
        if o.stdout:
            self.pdfPath = sys.stdout
        else:
            if srcPath != sys.stdin:
                self.pdfPath = os.path.splitext(srcPath)[0] + '.pdf'
            else:
                self.pdfPath = sys.stdout


    def beginDocument(self):
        """Things to do when a new document should be started.

        The initial page counter is 0, meaning that beginPage()
        will be called by beginLine()...
        """

        # Set initial page number and store file name.
        self.pageNum = 0
        o = self.options

        if not o.multiPage:
            # Create canvas.
            size = o.realPaperFormat.size
            self.canvas = canvas.Canvas(self.pdfPath, size, verbosity=0)
            c = self.canvas
            c.setPageCompression(1)
            c.setFont(o.fontName, o.fontSize)

            # Create document meta information.
            self.setPDFMetaInfo()

            # Set drawing frame.
            self.setMainFrame()

            # Determine the left text margin by adding the with
            # of the line number to the left margin of the main frame.
            format = "%%%dd " % len(`self.numLines`)
            fn, fs = self.currFont
            text = format % self.lineNum
            tm, bm, lm, rm = self.frame
            self.txm = lm
            if o.lineNum:
                self.txm = self.txm + c.stringWidth(text, fn, fs)


    def beginPage(self):
        "Things to do when a new page has to be added."

        o = self.options
        self.pageNum = self.pageNum + 1

        if not o.multiPage:
            tm, bm, lm, rm = self.frame
            self.text = self.canvas.beginText(lm, tm - o.fontSize)
            self.setFillColorAndFont(self.currColor, self.currFont)
        else:
            # Fail if stdout desired (with multiPage).
            if o.stdout:
                raise "IOError", "Can't create multiple pages on stdout!"

            # Create canvas with a modified path name.
            base, ext = os.path.splitext(self.pdfPath)
            newPath = "%s-%d%s" % (base, self.pageNum, ext)
            size = o.realPaperFormat.size
            self.canvas = canvas.Canvas(newPath, size, verbosity=0)
            c = self.canvas
            c.setPageCompression(1)
            c.setFont(o.fontName, o.fontSize)

            # Create document meta information.
            self.setPDFMetaInfo()

            # Set drawing frame.
            self.setMainFrame()

            tm, bm, lm, rm = self.frame
            self.text = self.canvas.beginText(lm, tm - o.fontSize)
            self.setFillColorAndFont(self.currColor, self.currFont)

        self.putPageDecoration()


    def beginLine(self, wrapped=0):
        "Things to do when a new line has to be added."

        # If there is no page yet, create the first one.
        if self.pageNum == 0:
            self.beginPage()

        # If bottom of current page reached, do a page break.
        # (This works only with one text object being used
        # for the entire page. Otherwise we need to maintain
        # the vertical position of the current line ourself.)
        y = self.text.getY()
        tm, bm, lm, rm = self.frame
        if y < bm:
            self.endPage()
            self.beginPage()

        # Print line number label, if needed.
        o = self.options
        if o.lineNum:
            #self.putLineNumLabel()
            font = ('Courier', o.fontSize)

            if not wrapped:
                # Print a label containing the line number.
                self.setFillColorAndFont(o.restCol, font)
                format = "%%%dd " % len(`self.numLines`)
                self.text.textOut(format % self.lineNum)
            else:
                # Print an empty label (using bgCol). Hackish!
                currCol = self.currColor
                currFont = self.currFont
                self.setFillColorAndFont(o.bgCol, font)
                self.text.textOut(' '*(len(`self.numLines`) + 1))
                self.setFillColorAndFont(currCol, currFont)


    def endLine(self, wrapped=0):
        "Things to do after a line is basically done."

        # End the current line by adding an 'end of line'.
        # (Actually done by the text object...)
        self.text.textLine('')

        if not wrapped:
            self.lineNum = self.lineNum + 1


    def endPage(self):
        "Things to do after a page is basically done."

        c = self.canvas

        # Draw the current text object (later we might want
        # to do that after each line...).
        c.drawText(self.text)
        c.showPage()

        if self.options.multiPage:
            c.save()


    def endDocument(self):
        "Things to do after the document is basically done."

        c = self.canvas

        # Display rest of last page and save it.
        c.drawText(self.text)
        c.showPage()
        c.save()


    def end(self):
        "Things to do after everything has been done."

        pass


    ### The real meat: methods writing something to a canvas.

    def putLineNumLabel(self, text, wrapped=0):
        "Add a long text that can't be split into chunks."

        o = self.options
        font = ('Courier', o.fontSize)

        if not wrapped:
            # Print a label containing the line number.
            self.setFillColorAndFont(o.restCol, font)
            format = "%%%dd " % len(`self.numLines`)
            self.text.textOut(format % self.lineNum)
        else:
            # Print an empty label (using bgCol). Hackish!
            currCol = self.currColor
            currFont = self.currFont
            self.setFillColorAndFont(o.bgCol, font)
            self.text.textOut(' '*(len(`self.numLines`) + 1))
            self.setFillColorAndFont(currCol, currFont)


    # Tried this recursively before, in order to determine
    # an appropriate string limit rapidly, but this wasted
    # much space and showed very poor results...
    # This linear method is very slow, but such lines should
    # be very rare, too!

    def putSplitLongText(self, text):
        "Add a long text that can't be split into chunks."

        # Now, the splitting will be with 'no mercy',
        # at the right margin of the main drawing area.

        M = len(text)
        t = self.text
        x = t.getX()
        o = self.options
        tm, bm, lm, rm = self.frame

        width = self.canvas.stringWidth
        fn, fs = self.currFont
        tw = width(text, fn, fs)

        tx = self.text
        if tw > rm - lm - x:
            i = 1
            T = ''
            while text:
                T = text[:i]
                tx = self.text # Can change after a page break.
                tw = width(T, fn, fs)

                if x + tw > rm:
                    tx.textOut(T[:-1])
                    self.endLine(wrapped=1)
                    self.beginLine(wrapped=1)
                    x = tx.getX()
                    text = text[i-1:]
                    M = len(text)
                    i = 0

                i = i + 1

                if i > M:
                    break

            tx.textOut(T)

        else:
            t.textOut(text)


    def putLongText(self, text):
        "Add a long text by gracefully splitting it into chunks."

        # Splitting is currently done only at blanks, but other
        # characters such as '.' or braces are also good
        # possibilities... later...

        o = self.options
        tm, bm, lm, rm = self.frame

        width = self.canvas.stringWidth
        fn, fs = self.currFont
        tw = width(text, fn, fs)

        arr = string.split(text, ' ')

        for i in range(len(arr)):
            a = arr[i]
            t = self.text # Can change during the loop...
            tw = width(a, fn, fs)
            x = t.getX()

            # If current item does not fit on current line, have it
            # split and then put before/after a line (maybe also
            # page) break.
            if x + tw > rm:
                self.putSplitLongText(a)
                t = self.text # Can change after a page break...

            # If it fits, just add it to the current text object.
            else:
                t.textOut(a)

            # Add the character we used to split th original text.
            if i < len(arr) - 1:
                t.textOut(' ')


    def putText(self, text):
        "Add some text to the current line."

        t = self.text
        x = t.getX()
        o = self.options
        fn, fs = o.fontName, o.fontSize
        tw = self.canvas.stringWidth(text, fn, fs)
        rm = self.frame[3]

        if x + tw < rm:
            t.textOut(text)
        else:
            self.putLongText(text)


    # Not yet tested.
    def putLine(self, text):
        "Add a line to the current text."

        self.putText(text)
        self.endLine()


    def putPageDecoration(self):
        "Draw some decoration on each page."

        # Use some abbreviations.
        o = self.options
        c = self.canvas
        tm, bm, lm, rm = self.frame

        # Restore default font.
        c.setFont(o.fontName, o.fontSize)

        c.setLineWidth(0.5) # in pt.

        # Background color.
        c.setFillColor(o.bgCol)
        pf = o.realPaperFormat.size
        c.rect(0, 0, pf[0], pf[1], stroke=0, fill=1)

        # Header.
        c.setFillColorRGB(0, 0, 0)
        c.line(lm, tm + .5*cm, rm, tm + .5*cm)
        c.setFont('Times-Italic', 12)

        if self.pdfPath == sys.stdout:
            filename = o.title or ' '
        else:
            path = os.path.basename(self.srcPath)
            filename = o.title or path

        c.drawString(lm, tm + 0.75*cm + 2, filename)

        # Footer.
        c.line(lm, bm - .5*cm, rm, bm - .5*cm)
        c.drawCentredString(0.5 * pf[0], 0.5*bm, "Page %d" % self.pageNum)

        # Box around main frame.
        # c.rect(lm, bm, rm - lm, tm - bm)


class PythonPDFLayouter (PDFLayouter):
    """A class to layout a simple multi-page PDF document.
    """

    ### API for adding specific Python entities.

    def addKw(self, t):
        "Add a keyword."

        o = self.options

        # Make base font bold.
        fam, b, i = fonts.ps2tt(o.fontName)
        ps = fonts.tt2ps(fam, 1, i)
        font = (ps, o.fontSize)

        self.setFillColorAndFont(o.kwCol, font)

        # Do bookmarking...
        if not o.noOutline and not o.multiPage:
            if t in ('class', 'def'):
                tm, bm, lm, rm = self.frame
                pos = self.text.getX()

                if pos == self.txm:
                    self.startPositions = []

                self.startPos = pos

                if not hasattr(self, 'startPositions'):
                    self.startPositions = []

                if pos not in self.startPositions:
                    self.startPositions.append(pos)

                # Memorize certain keywords.
                self.itemFound = t

            else:
                self.itemFound = None
                self.startPos = None

        self.putText(t)


    def addIdent(self, t):
        "Add an identifier."

        o = self.options

        # Make base font bold.
        fam, b, i = fonts.ps2tt(o.fontName)
        ps = fonts.tt2ps(fam, 1, i)
        font = (ps, o.fontSize)

        self.setFillColorAndFont(o.identCol, font)
        self.putText(t)

        # Bookmark certain identifiers (class and function names).
        if not o.noOutline and not o.multiPage:
            item = self.itemFound
            if item:
                # Add line height to current vert. position.
                pos = self.text.getY() + o.fontSize

                nameTag = "p%sy%s" % (self.pageNum, pos)
                c = self.canvas
                i = self.startPositions.index(self.startPos)
                c.bookmarkHorizontalAbsolute(nameTag, pos)
                c.addOutlineEntry('%s %s' % (item, t), nameTag, i)


    def addParam(self, t):
        "Add a parameter."

        o = self.options
        font = (o.fontName, o.fontSize)
        self.setFillColorAndFont(o.paramCol, font)
        self.text.putText(t)


    def addSimpleString(self, t):
        "Add a simple string."

        o = self.options
        font = (o.fontName, o.fontSize)
        self.setFillColorAndFont(o.strngCol, font)
        self.putText(t)


    def addTripleStringBegin(self):
        "Memorize begin of a multi-line string."

        # Memorise that we started a multi-line string.
        self.multiLineStringStarted = 1


    def addTripleStringEnd(self, t):
        "Add a multi-line string."

        self.putText(t)

        # Forget about the multi-line string again.
        self.multiLineStringStarted = 0


    def addComm(self, t):
        "Add a comment."

        o = self.options

        # Make base font slanted.
        fam, b, i = fonts.ps2tt(o.fontName)
        ps = fonts.tt2ps(fam, b, 1)
        font = (ps, o.fontSize)

        self.setFillColorAndFont(o.commCol, font)
        self.putText(t)


    def addRest(self, line, eol):
        "Add a regular thing."

        o = self.options

        # Nothing else to be done, print line as-is...
        if line:
            font = (o.fontName, o.fontSize)
            self.setFillColorAndFont(o.restCol, font)

            # ... except if the multi-line-string flag is set, then we
            # decide to change the current color to that of strings and
            # just go on.
            if self.multiLineStringStarted:
                self.setFillColorAndFont(o.strngCol, font)

            self.putText(line)

        # Print an empty line.
        else:
            if eol != -1:
                self.putText('')

    ### End of API.


class EmptyPythonPDFLayouter (PythonPDFLayouter):
    """A PDF layout with no decoration and no margins.

    The main frame extends fully to all paper edges. This is
    useful for creating PDFs when writing one page per file,
    in order to provide pre-rendered, embellished Python
    source code to magazine publishers, who can include the
    individual files and only need to add their own captures.
    """

    def setMainFrame(self, frame=None):
        "Make a frame extending to all paper edges."

        width, height = self.options.realPaperFormat.size
        self.frame = height, 0, 0, width


    def putPageDecoration(self):
        "Draw no decoration at all."

        pass


### Pretty-printing classes.

class PDFPrinter:
    """Generic PDF Printer class.

    Does not do much, but write a PDF file created from
    any ASCII input file.
    """

    outFileExt = '.pdf'


    def __init__(self, options=None):
        "Initialisation."

        self.data = None     # Contains the input file.
        self.inPath = None   # Path of input file.
        self.Layouter = PDFLayouter

        if type(options) != type(None):
            self.options = options
        else:
            self.options = Options()


    ### I/O.

    def readFile(self, path):
        "Read the content of a file."

        if path == sys.stdin:
            f = path
        else:
            f = open(path)

        self.inPath = path

        data = f.read()
        o = self.options
        self.data = re.sub('\t', ' '*o.tabSize, data)
        f.close()


    def formatLine(self, line, eol=0):
        "Format one line of Python source code."

        font = ('Courier', 8)
        self.layouter.setFillColorAndFont(Color(0, 0, 0), font)
        self.layouter.putText(line)


    def writeData(self, srcCodeLines, inPath, outPath=None):
        "Convert Python source code lines into a PDF document."

        # Create a layouter object.
        self.layouter = self.Layouter(self.options)
        l = self.layouter

        # Loop over all tagged source lines, dissect them into
        # Python entities ourself and let the layouter do the
        # rendering.
        splitCodeLines = string.split(srcCodeLines, '\n')

        ### Must also handle the case of outPath being sys.stdout!!
        l.begin(inPath, len(splitCodeLines))
        l.beginDocument()

        for line in splitCodeLines:
            l.beginLine()
            self.formatLine(line)
            l.endLine()

        l.endDocument()
        l.end()


    def writeFile(self, data, inPath=None, outPath=None):
        "Write some data into a file."

        if inPath == sys.stdin:
            self.outPath = sys.stdout
        else:
            if not outPath:
                path = os.path.splitext(self.inPath)[0]
                self.outPath = path + self.outFileExt

        self.writeData(data, inPath, outPath or self.outPath)


    def process(self, inPath, outPath=None):
        "The real 'action point' for working with Pretty-Printers."

        self.readFile(inPath)
        self.writeFile(self.data, inPath, outPath)


class PythonPDFPrinter (PDFPrinter):
    """A class to nicely format tagged Python source code.

    """

    comm  = 'COMMENT'
    kw    = 'KEYWORD'
    strng = 'STRING'
    ident = 'IDENT'
    param = 'PARAMETER'

    outFileExt = '.pdf'


    def __init__(self, options=None):
        "Initialisation, calling self._didInit() at the end."

        if type(options) != type(None):
            self.options = options
        else:
            self.options = Options()

        self._didInit()


    def _didInit(self):
        "Post-Initialising"

        # Define regular expression patterns.
        s = self
        comp = re.compile

        s.commPat   = comp('(.*?)<'  + s.comm  + '>(.*?)</' + s.comm  + '>(.*)')
        s.kwPat     = comp('(.*?)<'  + s.kw    + '>(.*?)</' + s.kw    + '>(.*)')
        s.identPat  = comp('(.*?)<'  + s.ident + '>(.*?)</' + s.ident + '>(.*)')
        s.paramPat  = comp('(.*?)<'  + s.param + '>(.*?)</' + s.param + '>(.*)')
        s.stPat     = comp('(.*?)<'  + s.strng + '>(.*?)</' + s.strng + '>(.*)')
        s.strng1Pat = comp('(.*?)<'  + s.strng + '>(.*)')
        s.strng2Pat = comp('(.*?)</' + s.strng + '>(.*)')
        s.allPat    = comp('(.*)')

        cMatch  = s.commPat.match
        kMatch  = s.kwPat.match
        iMatch  = s.identPat.match
        pMatch  = s.paramPat.match
        sMatch  = s.stPat.match
        s1Match = s.strng1Pat.match
        s2Match = s.strng2Pat.match
        aMatch  = s.allPat.match

        self.matchList = ((cMatch, 'Comm'),
            (kMatch, 'Kw'),
            (iMatch, 'Ident'),
            (pMatch, 'Param'),
            (sMatch, 'SimpleString'),
            (s1Match, 'TripleStringBegin'),
            (s2Match, 'TripleStringEnd'),
            (aMatch, 'Rest'))

        self.Layouter = PythonPDFLayouter

        # Load fontifier.
        self.tagFunc = loadFontifier(self.options)


    ###

    def formatLine(self, line, eol=0):
        "Format one line of Python source code."

        # Values for eol: -1:no-eol, 0:dunno-yet, 1:do-eol.

        for match, meth in self.matchList:
            res = match(line)

            if res:
                groups = res.groups()
                method = getattr(self, '_format%s' % meth)
                method(groups, eol)
                break


    def _formatIdent(self, groups, eol):
        "Format a Python identifier."

        before, id, after = groups
        self.formatLine(before, -1)
        self.layouter.addIdent(id)
        self.formatLine(after, eol)


    def _formatParam(self, groups, eol):
        "Format a Python parameter."

        before, param, after = groups
        self.formatLine(before, -1)
        self.layouter.addParam(before)
        self.formatLine(after, eol)


    def _formatSimpleString(self, groups, eol):
        "Format a Python one-line string."

        before, s, after = groups
        self.formatLine(before, -1)
        self.layouter.addSimpleString(s)
        self.formatLine(after, eol)


    def _formatTripleStringBegin(self, groups, eol):
        "Format a Python multi-line line string (1)."

        before, after = groups
        self.formatLine(before, -1)
        self.layouter.addTripleStringBegin()
        self.formatLine(after, 1)


    def _formatTripleStringEnd(self, groups, eol):
        "Format a Python multi-line line string (2)."

        before, after = groups
        self.layouter.addTripleStringEnd(before)
        self.formatLine(after, 1)


    def _formatKw(self, groups, eol):
        "Format a Python keyword."

        before, kw, after = groups
        self.formatLine(before, -1)
        self.layouter.addKw(kw)
        self.formatLine(after, eol)


    def _formatComm(self, groups, eol):
        "Format a Python comment."

        before, comment, after = groups
        self.formatLine(before, -1)
        self.layouter.addComm(comment)
        self.formatLine(after, 1)


    def _formatRest(self, groups, eol):
        "Format a piece of a Python line w/o anything special."

        line = groups[0]
        self.layouter.addRest(line, eol)


    ###

    def writeData(self, srcCodeLines, inPath, outPath=None):
        "Convert Python source code lines into a PDF document."

        # Create a layouter object.
        self.layouter = self.Layouter(self.options)
        l = self.layouter

        # Loop over all tagged source lines, dissect them into
        # Python entities ourself and let the layouter do the
        # rendering.
        splitCodeLines = string.split(srcCodeLines, '\n')
        l.begin(inPath, len(splitCodeLines))
        l.beginDocument()

        for line in splitCodeLines:
            l.beginLine()
            self.formatLine(line)
            l.endLine()

        l.endDocument()
        l.end()


    def process(self, inPath, outPath=None):
        "The real 'action point' for working with Pretty-Printers."

        self.readFile(inPath)
        self.taggedData = self._fontify(self.data)
        self.writeFile(self.taggedData, inPath, outPath)


    ### Fontifying.

    def _fontify(self, pytext):
        ""

        formats = {
            'rest'       : ('', ''),
            'comment'    : ('<%s>' % self.comm,  '</%s>' % self.comm),
            'keyword'    : ('<%s>' % self.kw,    '</%s>' % self.kw),
            'parameter'  : ('<%s>' % self.param, '</%s>' % self.param),
            'identifier' : ('<%s>' % self.ident, '</%s>' % self.ident),
            'string'     : ('<%s>' % self.strng, '</%s>' % self.strng) }

        # Parse.
        taglist = self.tagFunc(pytext)

        # Prepend special 'rest' tag.
        taglist[:0] = [('rest', 0, len(pytext), None)]

        # Prepare splitting.
        splits = []
        self._addSplits(splits, pytext, formats, taglist)

        # Do splitting & inserting.
        splits.sort()
        l = []
        li = 0

        for ri, dummy, insert in splits:
            if ri > li:
                l.append(pytext[li:ri])

            l.append(insert)
            li = ri

        if li < len(pytext):
            l.append(pytext[li:])

        return string.join(l, '')


    def _addSplits(self, splits, text, formats, taglist):
        ""

        # Helper for fontify().
        for id, left, right, sublist in taglist:

            try:
                pre, post = formats[id]
            except KeyError:
                # msg = 'Warning: no format '
                # msg = msg + 'for %s specified\n'%repr(id)
                # sys.stderr.write(msg)
                pre, post = '', ''

            if type(pre) != type(''):
                pre = pre(text[left:right])

            if type(post) != type(''):
                post = post(text[left:right])

            # len(splits) is a dummy used to make sorting stable.
            splits.append((left, len(splits), pre))

            if sublist:
                self._addSplits(splits, text, formats, sublist)

            splits.append((right, len(splits), post))


### Main

def main(cmdline):
    "Process command line as if it were sys.argv"

    # Create default options and initialize with argv
    # from the command line.
    options = Options()
    options.updateWithContentsOfArgv(cmdline[1:])

    # Print help message if desired, then exit.
    if options.h or options.help:
        print __doc__
        sys.exit()

    # Apply modest consistency checks and exit if needed.
    cmdStr = string.join(cmdline, ' ')
    find = string.find
    if find(cmdStr, 'paperSize') >= 0 and find(cmdStr, 'paperFormat') >= 0:
        details = "You can specify either paperSize or paperFormat, "
        details = detail + "but not both!"
        raise 'ValueError', details

    # Create PDF converter and pass options to it.
    if options.input:
        input = string.lower(options.input)

        if input == 'python':
            P = PythonPDFPrinter
        elif input == 'ascii':
            P = PDFPrinter
        else:
            details = "Input file type must be 'python' or 'ascii'."
            raise 'ValueError', details

    else:
        P = PythonPDFPrinter

    p = P(options)

    # Display options if needed.
    if options.v or options.verbose:
        pass # p.options.display()

    # Start working.
    verbose = options.v or options.verbose

    if options.stdout:
        if len(options.files) > 1 and verbose:
            print "Warning: will only convert first file on command line."
        f = options.files[0]
        p.process(f, sys.stdout)
    else:
        if verbose:
            print 'py2pdf: working on:'

        for f in options.files:
            try:
                if verbose:
                    print '  %s' % f
                if f != '-':
                    p.process(f)
                else:
                    p.process(sys.stdin, sys.stdout)
            except IOError:
                if verbose:
                    print '(IOError!)',

        if verbose:
            print
            print 'Done.'


def process(*args, **kwargs):
    "Provides a way of using py2pdf from within other Python scripts."

    noValOpts = 'h v verbose landscape stdout lineNum marcs help multiPage noOutline'
    noValOpts = string.split(noValOpts, ' ')

    s = 'py2pdf.py'
    for k, v in kwargs.items():
        if len(k) == 1:
            s = s + ' -%s' % k
            if k not in noValOpts:
                s = s + ' %s' % v
        elif len(k) > 1:
            s = s + ' --%s' % k
            if k not in noValOpts:
                s = s + '=%s' % v
    s = s + ' ' + string.join(args, ' ')
    s = string.split(s, ' ')

    main(s)


###

if __name__=='__main__': #NORUNTESTS
    main(sys.argv)